@@ -60,7 +60,7 @@ impl PrecompressOptimizer<'_> {
             });
 
             if has_decl {
-                let data = Some(analyze(&*stmts, self.marks));
+                let data = Some(analyze(&*stmts, Some(self.marks)));
 
                 stmts.visit_mut_children_with(&mut PrecompressOptimizer {
                     options: self.options,
